#ifndef BLD_CONTEXT_H
#define BLD_CONTEXT_H

/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <gpl/gpl.h>
#include <bld/scope.h>

/***********************************************************************
class BLD_Context

***********************************************************************/
class BLD_Context
{
	public:
		static const GPL_String rootScopeName;
		static const GPL_String scopeDelimiter;

		BLD_Context(void);
		virtual ~BLD_Context(void);

		GPL_Boolean AddRule(const GPL_String & string, const BLD_RuleType type,
							const OpArray & ops);
		BLD_BuildStatus Build(const GPL_String & name);
		BLD_BuildStatus Check(const GPL_String & name,
								const GPL_Boolean descend = TRUE);
		GPL_Boolean CreateTarget(const GPL_String & name, BLD_Target & target);
		GPL_Boolean Pop(BLD_Scope & scope);
		GPL_Boolean Push(const BLD_Scope & scope);

		void SetValue(const GPL_String & name, const BLD_Value & value);
		void SetValue(const GPL_String & scope, const GPL_String & variable,
						const BLD_Value & value);
		BLD_Value GetValue(const GPL_String & name) const;
		BLD_Value GetValue(const GPL_String & scope,
							const GPL_String & variable) const;
		void GetValues(GPL_Map<GPL_String, BLD_Value> & values) const;
		GPL_String Evaluate(const GPL_String & text) const;

	protected:
		static GPL_RegularExpression internalTargetPattern;
		GPL_Stack<BLD_Scope> scopes;
};

#endif /* BLD_CONTEXT_H */
