
/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <bld/environ.h>

/***********************************************************************
class BLD_Environment
***********************************************************************/
BLD_Environment *BLD_Environment::singleton = (BLD_Environment *) GPL_BAD_ADDRESS;

BLD_Environment::BLD_Environment(void)
{
}

BLD_Environment::~BLD_Environment(void)
{
}

BLD_Environment *BLD_Environment::GetEnvironment(void)
{
	if ((BLD_Environment *) GPL_BAD_ADDRESS == singleton)
	{
		singleton = new BLD_Environment();
	}
	return singleton;
}

GPL_Boolean BLD_Environment::GetValue(const GPL_String & name,
										GPL_String & value) const
{
	register char *t;
	for (register char **s = envp; (char **) 0 != s && (char *) 0 != (*s); ++s)
	{
		if ((char *) 0 != (t = strchr((*s), '=')) &&
			0 == strncmp(name.GetData(), (*s), t - (*s)))
		{
			value = ++t;
			return TRUE;
		}
	}
	return FALSE;
}

GPL_Boolean BLD_Environment::Execute(const GPL_String & command,
										GPL_String & output,
										GPL_Boolean inlinePrint) const
{
	// If there is a command...
	GPL_Boolean success = TRUE;
	output.Clear();
	char *str = (char *) 0;
	if (0 < command.GetLength())
	{
		// Execute the command.
		int status = 0;
		success = osdExecute(command.GetData(), &status, &str, (int)inlinePrint) && 0 == status;

		// If the command produced output...
		if ((char *) 0 != str)
		{
			// If the output has contents...
			register const unsigned long n = strlen(str);
			if (0 < n)
			{
				// Strip trailing new-lines.
				register unsigned long i = n - 1;
				while (0 <= i && '\n' == str[i])
				{
					--i;
				}
				output.SetData(str, i + 1);
			}
			osdFree(str);
		}
	}
	return success;
}

GPL_Boolean BLD_Environment::Execute(const GPL_String & program,
										const GPL_Array<GPL_String> & arguments,
										GPL_String & output,
										GPL_Boolean inlinePrint) const
{
	GPL_String command(1024, 0);
	command.Append(program);
	register const unsigned long n = arguments.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		command.Join(" ", arguments[i]);
	}
	if (!Execute(command, output,inlinePrint))
	{
		return FALSE;
	}
	return TRUE;
}

GPL_Array<GPL_String> BLD_Environment::Split(const GPL_String & text)
{
	GPL_String string;
	GPL_Array<GPL_String> strings;
	GPL_Boolean isSpace = FALSE;
	register const char *s = text.GetData();
	register const unsigned long n = text.GetLength();
	unsigned long start = 0;
	register unsigned long i = 0;
	for (i = 0; n > i; ++i)
	{
		isSpace = (' ' == s[i] || '\t' == s[i] || '\n' == s[i]);

		// If this is an escaped character...
		if ('\\' == s[i])
		{
			// If there are non-space characters preceding this...
			if (start < i)
			{
				string.Append(text.GetSubString(start, i - 1));
			}

			// Append the appended character.
			string.Append(s[++i]);
			start = i + 1;
		}

		// If this is a space character...
		else if (isSpace)
		{
			// If there are non-space characters preceding this...
			if (start < i)
			{
				string.Append(text.GetSubString(start, i - 1));
				strings.AppendElement(string);
				string.Clear();
			}
			start = i + 1;
		}
	}

	// If there are non-space characters preceding this...
	if (start < i)
	{
		string.Append(text.GetSubString(start, i - 1));
		strings.AppendElement(string);
	}
	return strings;
}
