#ifndef BLD_STATE_H
#define BLD_STATE_H

/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <gpl/gpl.h>

/***********************************************************************
class BLD_State

***********************************************************************/
class BLD_State : public GPL_Depictable
{
	public:
		BLD_State(const unsigned long f = 0) : flags(f) {}
		virtual ~BLD_State(void) {}

		GPL_Boolean operator==(const BLD_State & s) const
		{
			return flags == s.flags && lastBuilt == s.lastBuilt;
 		}

		inline GPL_Boolean HasTimeStamp(void) const { return flags & BLD_STATE_HAS_TIMESTAMP; }
		inline GPL_Boolean GetChecked(void) const { return flags & BLD_STATE_CHECKED; }
		inline GPL_Boolean GetExecuted(void) const { return flags & BLD_STATE_EXECUTED_COMMANDS; }
		inline GPL_Boolean GetFailed(void) const { return flags & BLD_STATE_BUILD_FAILED; }
		inline const GPL_Time & GetLastBuilt(void) const { return lastBuilt; }

		inline void SetChecked(void) { flags = flags | BLD_STATE_CHECKED; }
		inline void SetExecuted(void) { flags = flags | BLD_STATE_EXECUTED_COMMANDS; }
		inline void SetFailed(void) { flags = flags | BLD_STATE_BUILD_FAILED; }
		void SetLastBuilt(void);
		void SetLastBuilt(const GPL_Time & t);

		// GPL_Depictable
		virtual GPL_Boolean Print(GPL_String & string,
							const GPL_Verbosity verbosity) const;
		virtual GPL_Boolean Parse(const GPL_String & string);

	protected:
		enum Flags
		{
			// !(flags & BLD_STATE_HAS_TIMESTAMP) indicates that the
			// target has no corresponding file from which to get a
			// default time stamp and has never been built...
			BLD_STATE_HAS_TIMESTAMP		=	0x01,

			// !(flags & BLD_STATE_CHECKED) indicates that the
			// target has never been named in BLD_Context::Check().
			BLD_STATE_CHECKED			=	0x02,

			// !(flags & BLD_STATE_EXECUTED_COMMANDS) indicates that the
			// target has not executed any commands to achieve its
			// current state.
			BLD_STATE_EXECUTED_COMMANDS	=	0x04,

			// flags & BLD_STATE_BUILD_FAILED indicates that the
			// target failed its attempt to build itself.
			BLD_STATE_BUILD_FAILED		=	0x08
		};

		unsigned long flags;
		GPL_Time lastBuilt;
};

#endif /* BLD_STATE_H */

