#ifndef BLD_VAROP_H
#define BLD_VAROP_H

/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <gpl/gpl.h>
#include <bld/op.h>

enum BLD_SetType
{
	BLD_SET_PREPEND		=	-1,
	BLD_SET_OVERWRITE	=	0,
	BLD_SET_APPEND		=	1
};

/***********************************************************************
class BLD_SetOperation

Set a variable.
***********************************************************************/
class BLD_SetOperation : public BLD_Operation
{
	public:
		BLD_SetOperation(const BLD_SetType t, const GPL_String & n,
							BLD_Operation *o) : type(t), name(n), op(o) {}
		virtual ~BLD_SetOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		BLD_SetType type;
		GPL_String name;
		GPL_Pointer<BLD_Operation> op;
};

/***********************************************************************
class BLD_GetOperation

Get a variable.
***********************************************************************/
class BLD_GetOperation : public BLD_Operation
{
	public:
		BLD_GetOperation(const GPL_String & n) : name(n) {}
		virtual ~BLD_GetOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		GPL_String name;
};

/***********************************************************************
class BLD_BeginScopeOperation

Begin a scope.
***********************************************************************/
class BLD_BeginScopeOperation : public BLD_Operation
{
	public:
		BLD_BeginScopeOperation(const GPL_String & n) : name(n) {}
		virtual ~BLD_BeginScopeOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		GPL_String name;
};

/***********************************************************************
class BLD_EndScopeOperation

End a scope.
***********************************************************************/
class BLD_EndScopeOperation : public BLD_Operation
{
	public:
		BLD_EndScopeOperation(void) {}
		virtual ~BLD_EndScopeOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;
};

#endif /* BLD_VAROP_H */
