
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <iws/iws.h>

/***********************************************************************
class IWS_Application
***********************************************************************/
EW_EventContext *IWS_Application::eventContext = NULL;

IWS_Application::IWS_Application() : EW_Application(this)
{
	if(IWS_APP == IWS_DEFAULT_APP) delete IWS_APP;
	IWS_APP = this;
	singleton = this;
}

IWS_Application::~IWS_Application()
{
}

GPL_Boolean IWS_Application::Initialize(void)
{
	SetEventContext(eventContext);

	if (!EW_Application::Initialize())
	{
		return FALSE;
	}

	long screenx, screeny;
	GetEventContext()->GetPhantom()->GetScreenSize(&screenx, &screeny);
	long fontsize;

	if(screeny >= 1280)
	{
		fontsize = 16;
	}
	else if(screeny >= 1024)
	{
		fontsize = 14;
	}
	else if(screeny >= 768)
	{
		fontsize = 12;
	}
	else if(screeny >= 600)
	{
		fontsize = 10;
	}
	else if(screeny >= 480)
	{
		fontsize = 8;
	}
	else
	{
		fontsize = 8;
	}

	// Load the font information.
	EW_FontRequest request;
	request.SetHeight(fontsize);
	request.SetStyle(EW_FONT_PROPORTIONAL);
	EW_Font font;
	if (!GetEventContext()->GetPhantom()->LoadFont(&font, &request))
	{
		GetEventContext()->GetPhantom()->SetFont(&font);
		GetEventContext()->GetPhantom()->CopyDefaultFontFrom(&font);
	}

	return TRUE;
}

void IWS_Application::Finalize(void)
{
	EW_Application::Finalize();
}

void IWS_Application::Process()
{
}

void IWS_Application::LoadConfiguration(const int argc, const char **argv, const char **envp)
{
	GPL_Aware gplAware;
	WBA_Aware wbaAware;
	EW_Aware ewAware;
	WDS_Aware wdsAware;
	IWS_Aware iwsAware;

	WBA_ENV.LoadEnvVariables(envp, FALSE);
	if (!WBA_ENV.LoadCommandLine(argc, argv, TRUE))
	{
		Printf("%s", GetUsage());
		SetExitCode(-1);
		Exit();
	}
}

void IWS_Application::GetApproximateSize(const char *string, long & width, long & height)
{
	if ((const char *) GPL_BAD_ADDRESS == string)
	{
		height = GetEventContext()->GetPhantom()->GetFontHeight();
		width = 0;
		return;
	}

	width = 0;
	height = 0;
	GPL_StringTokenizer tokens;
	tokens.Reset(string, "\n");
	GPL_String token = tokens.Next();
	while ((char *) 0 != token.GetData())
	{
		const long w = GetEventContext()->GetPhantom()->StringWidth(token) +
							(2 * GetEventContext()->GetPhantom()->CharWidth('W'));
		if (w > width)
		{
			width = w;
		}
		height += GetEventContext()->GetPhantom()->GetFontHeight();

		token = tokens.Next();
	}
}

/***********************************************************************
class IWS_ExitApplicationHandler
***********************************************************************/
void IWS_ExitApplicationHandler::HandleEvent(const GPL_String & string)
{
	IWS_APP->GetEventContext()->EndLoop();
}

/***********************************************************************
class IWS_LookShiftHandler
***********************************************************************/
void IWS_LookShiftHandler::HandleEvent(const GPL_String & string)
{
	//!! KLUDGE
}
