
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <iws/iws.h>

IWS_Page::IWS_Page()
{
}

IWS_Page::~IWS_Page()
{
}

void IWS_Page::Show()
{
	GPL_MapIterator<EW_Window *, GPL_Boolean> iterator;
	iterator.Reset (&windowState);
	while (iterator.Next ())
	{
		if (iterator.Value ())
		{
			IWS_Window::ShowWindow (iterator.Key ());
		}
	}
}


void IWS_Page::Hide()
{
	GPL_MapIterator<EW_Window *, GPL_Boolean> iterator;
	iterator.Reset (&windowState);
	while (iterator.Next ())
	{
		iterator.Value () = IWS_Window::IsWindowShown (iterator.Key ());
		IWS_Window::HideWindow (iterator.Key ());
	}
}

void IWS_Page::AddWindow (EW_Window * window, const GPL_Boolean shown) 
{
	windowState.Insert(window, shown);
}

void IWS_Page::RemoveWindow (EW_Window * window) 
{
	windowState.Remove (window);
}

GPL_Boolean IWS_Page::Print (GPL_String &string, const GPL_Verbosity verbosity) const 
{
	GPL_String image;
	GPL_MapIterator<EW_Window *, GPL_Boolean> iterator;
	iterator.Reset ((GPL_Map<EW_Window *, GPL_Boolean> *) &windowState);
	while (iterator.Next ()) 
	{
		image.Printf ("[%p,%s]", iterator.Key (),
						(iterator.Value () ? "SHOWN" : "HIDDEN"));
		string.Join (",", image);
	}
	return TRUE;
}

const GPL_String IWS_Book::defaultPage = "DEFAULT";

IWS_Book::IWS_Book (void) 
{
	// Start on the default page.
	currentPage = defaultPage;
}

IWS_Book::~IWS_Book (void) 
{
}

const GPL_String &IWS_Book::GetShownPage (void) const 
{
	return currentPage;
}

void IWS_Book::ShowPage (const GPL_String & page) 
{
	// Do nothing if this page is already being shown.
	if (page == currentPage) {

		return;
	}

	// Hide all the windows on the old page.
	pages[currentPage].Hide ();

	// Switch pages.
	currentPage = page;

	// Show all the windows on the new page.
	pages[currentPage].Show ();
}	

void IWS_Book::AddWindowToPage (EW_Window * window, const GPL_String & page,
									const GPL_Boolean shown) 
{
	// Ignore null windows.
	if ((EW_Window *) GPL_BAD_ADDRESS == window) 
	{
		return;
	}

	// Add the window to the page in the book.
	pages[page].AddWindow (window, shown);

	// If this page is the current page and the window should be shown...
	if (page == currentPage && shown) 
	{
		IWS_Window::ShowWindow (window);
	}
}

void IWS_Book::RemoveWindowFromPage (EW_Window * window, const GPL_String & page) {

	// Ignore null windows.
	if ((EW_Window *) GPL_BAD_ADDRESS == window) 
	{
		return;
	}

	// Hide the window no matter what.
	IWS_Window::HideWindow (window);

	// Remove the window from the book.
	pages[page].RemoveWindow (window);
}

