/******************************************************************************

Coral Tree iws/button.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <iws/button.h>
#include <iws/app.h>

/***********************************************************************
class IWS_Button
***********************************************************************/
IWS_DefaultButtonHandler IWS_Button::defaultHandler;

IWS_Button::IWS_Button(void) : image((WDS_Image *) GPL_BAD_ADDRESS)
{
	IWS_ButtonConstructor();
}

IWS_Button::IWS_Button(const GPL_String & n) : image((WDS_Image *) GPL_BAD_ADDRESS)
{
	IWS_ButtonConstructor();
	GPL_String s = n;
	SetLabel((char *)s);
	SetName((char *)s);
}

IWS_Button::~IWS_Button(void)
{
	delete image;
}

void IWS_Button::IWS_ButtonConstructor(void)
{
	SetGeneralFlags(WDS_WIDGET_FLAGS_AUTOSIZE);
	SetHandler(&defaultHandler);

	SetBounds(EW_HORIZONTAL, EW_BOUNDS_NONE);
	SetAnnotationMode(WDS_WIDGET_ANNOTATION_LEFT);
	SetTextSize(-1, -1);
	SetAnnotationSize(-1, -1);
	accelerator = -1;
}

void IWS_Button::SetEnabled(const GPL_Boolean e)
{
	SetGreyOut(!e);
	SetDormant(!e);
	SetDirty(EW_DIRTY_COMPLETE);
}

void IWS_Button::SetSizeMode(long xsetting, long ysetting)
{
	long gflags;
	gflags = WDS_WIDGET_FLAGS_NONE;

	switch(xsetting)
	{
		case IWS_SIZE_MODE_WRAP:
			gflags |= WDS_WIDGET_FLAGS_AUTOSIZE_X;
			SetBounds(EW_HORIZONTAL, EW_BOUNDS_NONE);
			break;
		case IWS_SIZE_MODE_FILL:
			SetBounds(EW_HORIZONTAL, EW_BOUNDS_FILL_PARENT);
			SetBounds(EW_HORIZONTAL, EW_BOUNDS_NONE);
			break;
	}

	switch(ysetting)
	{
		case IWS_SIZE_MODE_WRAP:
			gflags |= WDS_WIDGET_FLAGS_AUTOSIZE_Y;
			SetBounds(EW_VERTICAL, EW_BOUNDS_NONE);
			break;
		case IWS_SIZE_MODE_FILL:
			SetBounds(EW_VERTICAL, EW_BOUNDS_FILL_PARENT);
			SetBounds(EW_VERTICAL, EW_BOUNDS_NONE);
			break;
	}

	SetGeneralFlags(gflags);
}

void IWS_Button::SetHandler(IWS_Handler *handler, WDS_Activation *actv)
{
	if ((IWS_Handler *) GPL_BAD_ADDRESS == handler)
	{
		return;
	}

	WDS_Activation a;
	if ((WDS_Activation *) GPL_BAD_ADDRESS == actv)
	{
		actv = &a;
	}
	actv->SetObject((void *)handler);
	actv->SetTargetWidget(this);
	SetStateChangeCallback(WDS_STATE_0, handler->Callback, actv);
}

GPL_Boolean IWS_Button::SetImage(long index)
{
	return SetImage(wdsGetBuiltinXPM(index));
}

GPL_Boolean IWS_Button::SetImage(char **i)
{
	delete image;
	image = new WDS_Image ();
	if (!image->ReadFromBufferXPM (i))
	{
		delete image;
		image = (WDS_Image *) GPL_BAD_ADDRESS;
		return FALSE;
	}
	WDS_Widget::SetImage (WDS_STATE_ALL, image);
	SetLabel("");
	return TRUE;
}

GPL_Boolean IWS_Button::SetImageFromFile (const GPL_String & f)
{
	delete image;
	image = new WDS_Image ();
	if (!image->LoadFromFileXPM (f.GetData ())) 
	{
		delete image;
		image = (WDS_Image *) GPL_BAD_ADDRESS;
		return FALSE;
	}
	WDS_Widget::SetImage (WDS_STATE_ALL, image);
	SetLabel("");
	return TRUE;
}

const char *IWS_Button::GetText()
{
	return GetLabel();
}

void IWS_Button::SetAccelerator(long ascii)
{
	if(ascii == accelerator) return;
	EW_Event event;
	if(accelerator != -1 && ascii == -1)
	{
		event.SetSIS(EW_EVENT_KEYBOARD,EW_KEY_ALL,EW_STATE_ANY);
		RegisterEventBypass(FALSE,&event,WDS_BYPASS_FLAG_NULL,0);
	}
	event.SetSIS(EW_EVENT_KEYBOARD,EW_KEY_ALL,EW_STATE_ANY);
	RegisterEventBypass(TRUE,&event,WDS_BYPASS_FLAG_NULL,0);
	accelerator = ascii;
}

long IWS_Button::Operate(EW_Event *event)
{
	long return_val;

	if(	event->IsKey(EW_KEY_ALL) && 
		(event->GetWidgetFlags()&WDS_BYPASS_EVENT)
	  )
	{
		event->Print(TRUE);
		WDS_Scope *scope;
		scope = (WDS_Scope *)FindWidgetType(WDS_Scope::GetTypeCodeStatic(),FALSE);
		if(scope && !(scope->IsTaken(event)))
		{
			long item = event->GetItem();
			if(item == accelerator)
			{
				long toggleable= (GetFlags()&WDS_BUTTON_FLAGS_TOGGLE);
				if(!toggleable && event->GetState()&EW_STATE_RELEASE)
				{
					fprintf(stderr,"%c -\n", (char)item);
					ChangeState(WDS_STATE_0,event);
					SetDirty(EW_DIRTY_COMPLETE);
					event->SetUsed(TRUE);
				}
				else if(event->GetState()&EW_STATE_PRESS)
				{
					fprintf(stderr,"%c +\n", (char)item);
					if(toggleable)
					{
						if(GetWidgetState() && 
							!(GetFlags()&WDS_BUTTON_FLAGS_UNTAKE_ONLY) )
						{
							ChangeState(WDS_STATE_0,event);
						}
						else
						{
							ChangeState(WDS_STATE_1,event);
						}
					}
					else if(!GetWidgetState())
					{
						ChangeState(WDS_STATE_1,event);
					}
					SetDirty(EW_DIRTY_COMPLETE);	
					event->SetUsed(TRUE);
				}
			}
		}
		return_val = 0;
	}
	else
	{
		return_val = WDS_Button::Operate(event);
	}
	return return_val;
}

IWS_AnnotationAlignable_IMPLEMENTATION(IWS_Button, WDS_Button);
