
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <iws/iws.h>

IWS_DefaultRadioButtonSelectHandler 	IWS_RadioButton::defaultSelectHandler;
IWS_DefaultRadioButtonDeselectHandler	IWS_RadioButton::defaultDeselectHandler;
IWS_DefaultRadioButtonSelectHandler 	IWS_ToggleButton::defaultSelectHandler;
IWS_DefaultRadioButtonDeselectHandler	IWS_ToggleButton::defaultDeselectHandler;

IWS_RadioButton::IWS_RadioButton(const GPL_String &n)
{
	IWS_RadioButtonConstructor();
	GPL_String s = n;
	SetAnnotation((char *)s);
	SetName((char *)s);
}

IWS_RadioButton::IWS_RadioButton(void)
{
	IWS_RadioButtonConstructor();
}

IWS_RadioButton::~IWS_RadioButton(void)
{
	delete buttonImage;
}

void IWS_RadioButton::IWS_RadioButtonConstructor()
{
	buttonImage = (WDS_Image *) GPL_BAD_ADDRESS;
	SetSelectHandler(&defaultSelectHandler);
	SetDeselectHandler(&defaultDeselectHandler);
}

void IWS_RadioButton::SetSelectHandler(IWS_Handler *handler, WDS_Activation *actv)
{
	if(!handler) return;
	WDS_Activation a;
	if(!actv) actv = &a;
	actv->SetObject((void *)handler);
	SetStateChangeCallback(WDS_STATE_1, handler->Callback, actv);
}

void IWS_RadioButton::SetDeselectHandler(IWS_Handler *handler, WDS_Activation *actv)
{
	if(!handler) return;
	WDS_Activation a;
	if(!actv) actv = &a;
	actv->SetObject((void *)handler);
	SetStateChangeCallback(WDS_STATE_0, handler->Callback, actv);
}

GPL_Boolean IWS_RadioButton::SetImage(long index)
{
	return SetImage(wdsGetBuiltinXPM(index));
}

GPL_Boolean IWS_RadioButton::SetImage(char **i)
{
	buttonImage = new WDS_Image ();
	if (!buttonImage->ReadFromBufferXPM (i))
	{
		buttonImage = (WDS_Image *) GPL_BAD_ADDRESS;
		return FALSE;
	}
	WDS_Widget::SetImage (WDS_STATE_ALL, buttonImage);
	SetLabel("");
	return TRUE;
}

GPL_Boolean IWS_RadioButton::SetImageFromFile (const GPL_String & f) {

	buttonImage = new WDS_Image ();
	if (!buttonImage->LoadFromFileXPM (f.GetData ())) 
	{
		buttonImage = (WDS_Image *) GPL_BAD_ADDRESS;
		return FALSE;
	}
	WDS_Widget::SetImage (WDS_STATE_ALL, buttonImage);
	SetLabel("");
	return TRUE;
}

IWS_ToggleButton::IWS_ToggleButton(const GPL_String &n)
{
	IWS_ToggleButtonConstructor();
	GPL_String s = n;
	SetAnnotation((char *)s);
	SetName((char *)s);
}

IWS_ToggleButton::IWS_ToggleButton(void)
{
	IWS_ToggleButtonConstructor();
}

IWS_ToggleButton::~IWS_ToggleButton(void)
{
	delete buttonImage;
}

void IWS_ToggleButton::IWS_ToggleButtonConstructor()
{
	buttonImage = (WDS_Image *) GPL_BAD_ADDRESS;
	SetSelectHandler(&defaultSelectHandler);
	SetDeselectHandler(&defaultDeselectHandler);
}

void IWS_ToggleButton::SetSelectHandler(IWS_Handler *handler, WDS_Activation *actv)
{
	if(!handler) return;
	WDS_Activation a;
	if(!actv) actv = &a;
	actv->SetObject((void *)handler);
	SetStateChangeCallback(WDS_STATE_1, handler->Callback, actv);
}

void IWS_ToggleButton::SetDeselectHandler(IWS_Handler *handler, WDS_Activation *actv)
{
	if(!handler) return;
	WDS_Activation a;
	if(!actv) actv = &a;
	actv->SetObject((void *)handler);
	SetStateChangeCallback(WDS_STATE_0, handler->Callback, actv);
}

GPL_Boolean IWS_ToggleButton::SetImage(long index)
{
	return SetImage(wdsGetBuiltinXPM(index));
}

GPL_Boolean IWS_ToggleButton::SetImage(char **i)
{
	buttonImage = new WDS_Image ();
	if (!buttonImage->ReadFromBufferXPM (i))
	{
		buttonImage = (WDS_Image *) GPL_BAD_ADDRESS;
		return FALSE;
	}
	WDS_Widget::SetImage (WDS_STATE_ALL, buttonImage);
	SetLabel("");
	return TRUE;
}

GPL_Boolean IWS_ToggleButton::SetImageFromFile (const GPL_String & f) {

	buttonImage = new WDS_Image ();
	if (!buttonImage->LoadFromFileXPM (f.GetData ())) 
	{
		buttonImage = (WDS_Image *) GPL_BAD_ADDRESS;
		return FALSE;
	}
	WDS_Widget::SetImage (WDS_STATE_ALL, buttonImage);
	SetLabel("");
	return TRUE;
}

IWS_RadioBar::IWS_RadioBar()
{
	rootForm = new WDS_Form;
	AddWidget(rootForm);
	SetNodeSizing(EW_NODE_SIZING_FIT_CHILD);
	GetForm()->SetMode (WDS_FORM_ROW);
	GetForm()->SetUniformity (WDS_UNIFORM_BOTH);
	GetForm()->SetSizing (WDS_HORIZONTAL, WDS_SIZING_FIT_CHILDREN);
	GetForm()->SetSizing (WDS_VERTICAL, WDS_SIZING_FIT_CHILDREN);
}

IWS_RadioColumn::IWS_RadioColumn()
{
	GetForm()->SetMode (WDS_FORM_SINGLE_COLUMN);
	GetForm()->SetUniformity (WDS_UNIFORM_BOTH);
	GetForm()->SetSizing (WDS_HORIZONTAL, WDS_SIZING_FIT_CHILDREN);
	GetForm()->SetSizing (WDS_VERTICAL, WDS_SIZING_FIT_CHILDREN);
}

IWS_RadioBar::~IWS_RadioBar()
{
}

#define IWS_ADDBUTTON_BODY_HANDLE				\
	button->SetName (n.GetData ());				\
	button->SetAnnotation (n.GetData ());		\
												\
	if(sel)										\
	{											\
		button->SetSelectHandler(sel);			\
	}											\
	if(desel)									\
	{											\
		button->SetDeselectHandler(desel);		\
	}											\
	GetForm()->AddWidget(button);				\
	return button;

#define IWS_ADDBUTTON_BODY_RAW									\
	button->SetName 		(n.GetData ());						\
	button->SetAnnotation 	(n.GetData ());						\
																\
	if((void (*)(WDS_Activation *)) GPL_BAD_ADDRESS != sel)		\
	{															\
		button->SetStateChangeCallback(WDS_STATE_1, sel, a);	\
	}															\
	if((void (*)(WDS_Activation *)) GPL_BAD_ADDRESS != desel)	\
	{															\
		button->SetStateChangeCallback(WDS_STATE_0, desel, a);	\
	}															\
	GetForm()->AddWidget(button);								\
	return button;

IWS_RadioButton *IWS_RadioBar::AddRadio( const GPL_String &n, IWS_Handler *sel, IWS_Handler *desel)
{
	IWS_RadioButton *button = new IWS_RadioButton();
	IWS_ADDBUTTON_BODY_HANDLE
}

IWS_RadioButton *IWS_RadioBar::AddRadio( const GPL_String &n,
							void (*sel) (WDS_Activation *), 
							WDS_Activation * a,
							void (*desel) (WDS_Activation *))
{
	IWS_RadioButton *button = new IWS_RadioButton();
	IWS_ADDBUTTON_BODY_RAW
}

IWS_ToggleButton *IWS_RadioBar::AddToggle( const GPL_String &n, IWS_Handler *sel, IWS_Handler *desel)
{
	IWS_ToggleButton *button = new IWS_ToggleButton();
	IWS_ADDBUTTON_BODY_HANDLE
}

IWS_ToggleButton *IWS_RadioBar::AddToggle( const GPL_String &n,
							void (*sel) (WDS_Activation *), 
							WDS_Activation * a,
							void (*desel) (WDS_Activation *))
{
	IWS_ToggleButton *button = new IWS_ToggleButton();
	IWS_ADDBUTTON_BODY_RAW
}

WDS_Button *IWS_RadioBar::AddButton( WDS_Button *button)
{
	GetForm()->AddWidget(button);
	return button;
}

WDS_Button *IWS_RadioBar::FindButton(const GPL_String & n)
{
	EW_WidgetList *list = GetWidgetList();
	list->ToHead();
	/* ASSUMING AT LEAST A WDS_Widget */
	WDS_Widget *widget;
	while((WDS_Widget *) GPL_BAD_ADDRESS != (widget = (WDS_Widget *)(list->PostIncrement())))
	{
		if (widget->GetTypeCode() == WDS_Button::GetTypeCodeStatic())
		{
			if (n == widget->GetName())
			{
				return (WDS_Button *) widget;
			}
		}
	}
	return (WDS_Button *) GPL_BAD_ADDRESS;
}
