
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <mte/mte.h>

#if MTE_IMPL==MTE_SPROC
usptr_t *MTE_Mutex::arena = NULL;
#endif

MTE_Mutex::MTE_Mutex()
{
#if MTE_IMPL==MTE_SPROC
	if(!arena)
	{
		arena = usinit("./mutex_arena");
		if(!arena)
		{
			wbaLog("MTE ERROR", "sproc arena initialization failed");
		}
	}

	mutex = usnewlock(arena);
	if(!mutex)
	{
		wbaLog("MTE ERROR", "could not create sproc mutex");
	}
#elif MTE_IMPL==MTE_PTHREAD
	int n;
	n = pthread_mutex_init(&mutex, NULL);
	if(n)
	{
		wbaLog("MTE ERROR", "could not create pthread mutex");
	}
#elif MTE_IMPL==MTE_WIN32
	mutex = CreateMutex(NULL,FALSE,NULL);
	if(!mutex)
	{
		wbaLog("MTE ERROR", "could not create win32 mutex");
	}
#else
#endif
}

MTE_Mutex::MTE_Mutex(MTE_Mutex *m)
{
#if MTE_IMPL==MTE_SPROC
	arena = m->arena;
	mutex = m->mutex;
#elif MTE_IMPL==MTE_PTHREAD
	mutex = m->mutex;
#elif MTE_IMPL==MTE_WIN32
	mutex = m->mutex;
#else
#endif
}

MTE_Mutex::~MTE_Mutex()
{
#if MTE_IMPL==MTE_SPROC
#elif MTE_IMPL==MTE_PTHREAD
	int errcode;
	if(errcode = pthread_mutex_destroy(&mutex))
	{
		if(errcode == EBUSY)
		{
			Unlock();
			if(pthread_mutex_destroy(&mutex))
			{
				wbaLog("MTE ERROR", "could not destroy pthread mutex");
			}
		}
		else
		{
			wbaLog("MTE ERROR", "could not destroy pthread mutex");
		}
	}
#elif MTE_IMPL==MTE_WIN32
#else
#endif
}

GPL_Boolean MTE_Mutex::Lock()
{
#if MTE_IMPL==MTE_SPROC
	if(!ussetlock(mutex))
	{
		return FALSE;
	}
	return TRUE;
#elif MTE_IMPL==MTE_PTHREAD
	if(pthread_mutex_lock(&mutex))
	{
		return FALSE;
	}
	return TRUE;
#elif MTE_IMPL==MTE_WIN32
	if(WaitForSingleObject(mutex,INFINITE) != WAIT_OBJECT_0)
	{
		return FALSE;
	}
	return TRUE;
#else
	return TRUE;
#endif
}

GPL_Boolean MTE_Mutex::Unlock()
{
#if MTE_IMPL==MTE_SPROC
	if(usunsetlock(mutex))
	{
		return FALSE;
	}
	return TRUE;
#elif MTE_IMPL==MTE_PTHREAD
	if(pthread_mutex_unlock(&mutex))
	{
		return FALSE;
	}
	return TRUE;
#elif MTE_IMPL==MTE_WIN32
	if(!ReleaseMutex(mutex))
	{
		return FALSE;
	}
	return TRUE;
#else
	return TRUE;
#endif
}

