#ifndef SPA_MATRIX_H
#define SPA_MATRIX_H

/******************************************************************************

Coral Tree spa/matrix.h
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#define SPA_FAST

/******************************************************************************
class SPA_MatrixBase
******************************************************************************/
template <class T>
class SPA_MatrixBase
{
	public:
		GPL_Integer O;
		GPL_Integer P;

		virtual T& operator ()(const GPL_Integer m, const GPL_Integer n) = 0;
};

/******************************************************************************
class SPA_Matrix
******************************************************************************/
template <GPL_Integer M, GPL_Integer N, class T>
class SPA_Matrix : public SPA_MatrixBase<T>
{
	public:
		SPA_Matrix(void)
		{
			SPA_MatrixBase<T>::O=M;
			SPA_MatrixBase<T>::P=N;
			GPL_Integer i;
			GPL_Integer j;
			for (i = 0; i < M; ++i)
			{
				for (j = 0; j < N; ++j)
				{
					//data[i][j] = (T)i * 1000 + j; // for testing
					data[i][j] = 0;
				}
			}
		}
		virtual	~SPA_Matrix(void)
		{
		}

#ifdef SPA_SUBSCRIPT
		T* operator [](const GPL_Integer index)
		{
#ifdef SPA_RANGE_CHECKING
			if (index >= M)
			{
				wbaLog("SPA WARNING", "M out of range: %d is over %d", index, M - 1);
			}
#endif
			return (T *)(data[index]);
		}
#endif

		T& operator ()(const GPL_Integer m, const GPL_Integer n)
		{
			return Element(m,n);
		}

		SPA_Matrix<M,N,T> &operator =(const SPA_Matrix<M,N,T> &other)
		{
			GPL_Integer i;
			GPL_Integer j;
			for (i = 0; i < M; ++i)
			{
				for (j = 0; j < N; ++j)
				{
					data[i][j] = other.data[i][j];
				}
			}
			return *this;
		}


		T& Element(const GPL_Integer m, const GPL_Integer n)
		{
#ifdef SPA_RANGE_CHECKING
			if (m >= M)
			{
				wbaLog("SPA WARNING", "M out of range: %d is over %d", m, M - 1);
			}
			if (n >= N)
			{
				wbaLog("SPA WARNING", "N out of range: %d is over %d", n, n - 1);
			}
#endif
			return data[m][n];
		}

		void Clear(const T clrval = 0)
		{
			GPL_Integer i;
			GPL_Integer j;
			for (i = 0; i < M; ++i)
			{
				for (j = 0; j < N; ++j)
				{
					data[i][j] = clrval;
				}
			}
		}
		void Scale(const T factor)
		{
			GPL_Integer i;
			GPL_Integer j;
			for (i = 0; i < M; ++i)
			{
				for (j = 0; j < N; ++j)
				{
					data[i][j] *= factor;
				}
			}
		}
		void Scale(SPA_MatrixBase<T> *result, SPA_MatrixBase<T> *operand)
		{
			GPL_Integer i;
			GPL_Integer j;
			for (i = 0; i < M; i++)
			{
				for (j = 0; j < N; j++)
				{
					(*result)(i,j) = (*operand)(i,j) * data[i][j];
				}
			}
		}
		void Add(SPA_MatrixBase<T> *result, SPA_MatrixBase<T> *operand)
		{
			GPL_Integer i;
			GPL_Integer j;
			for (i = 0; i < M; i++)
			{
				for (j = 0; j < N; j++)
				{
					(*result)(i,j) = (*operand)(i,j) + data[i][j];
				}
			}
		}
		void Add(SPA_MatrixBase<T> &operand)
		{
			GPL_Integer i;
			GPL_Integer j;
			for (i = 0; i < M; i++)
			{
				for (j = 0; j < N; j++)
				{
					data[i][j] += operand(i,j);
				}
			}
		}
		void Subtract(SPA_MatrixBase<T> *result, SPA_MatrixBase<T> *operand)
		{
			GPL_Integer i;
			GPL_Integer j;
			for (i = 0; i < M; i++)
			{
				for (j = 0; j < N; j++)
				{
#ifdef SPA_FAST
					((SPA_Matrix<M,N,T> *)result)->data[i][j] = data[i][j] - ((SPA_Matrix<M,N,T> *)operand)->data[i][j];
#else
					(*result)(i,j) = data[i][j] - (*operand)(i,j);
#endif
				}
			}
		}
		void Subtract(SPA_MatrixBase<T> &operand)
		{
			GPL_Integer i;
			GPL_Integer j;
			for (i = 0; i < M; i++)
			{
				for (j = 0; j < N; j++)
				{
					data[i][j] -= operand(i,j);
				}
			}
		}
		void Print(GPL_String &string)
		{
			GPL_String tmp;
			GPL_Integer i;
			GPL_Integer j;
			string.Clear();
			for (i = 0; i < M; i++)
			{
				for (j = 0; j < N; j++)
				{
					tmp.Printf("%4.4f", data[i][j]);
					string.Append(tmp);
					string.Append(" ");
				}
				string.Append("\n");
			}
		}
		void Multiply(SPA_MatrixBase<T> *result, SPA_MatrixBase<T> *operand)
		{
			GPL_Integer i;
			GPL_Integer j;
			GPL_Integer k;
#ifdef SPA_RANGE_CHECKING
			if (result->O != M)
			{
				wbaLog("SPA WARNING", "result M != this M: %d != %d", result->O, M);
			}
			if (result->P != operand->P)
			{
				wbaLog("SPA WARNING", "result N != operand N: %d != %d", result->P, operand->P);
			}
			if (operand->O != N) 
			{
				wbaLog("SPA WARNING", "operand M != this N: %d != %d", operand->O, N);
			}
#endif
			for (i = 0; i < M; ++i)
			{
				for (j = 0; j < operand->P; ++j)
				{
					(*result)(i,j) = 0;
					for (k = 0; k < N; ++k)
					{
						(*result)(i,j) += data[i][k] * (*operand)(k,j);
					}
				}
			}
		}

	protected:
		T	data[M][N];
};

	

#endif /* SPA_MATRIX_H */
