#ifndef SPA_VECTOR_H
#define SPA_VECTOR_H

/******************************************************************************

Coral Tree spa/vector.h
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <spa/matrix.h>

/******************************************************************************
class SPA_Vector
******************************************************************************/
template <GPL_Integer S, class T>
class SPA_Vector
{
	public:
		virtual	T &operator [](const GPL_Integer index) = 0;
		virtual	T &Element(const GPL_Integer index) = 0;
		T Magnitude()
		{
			GPL_Integer i;
			T value = 0;
			if(S == 2)
			{
				T a,b;
				a = Element(0);
				b = Element(1);
				value = (T)sqrt((double)(a*a+b*b));
			}
			else if(S == 3)
			{
				T a,b,c;
				a = Element(0);
				b = Element(1);
				c = Element(2);
				value = (T)sqrt((double)(a*a+b*b+c*c));
			}
			else
			{
				for(i = 0;i < S;i++)
				{
					value += Element(i) * Element(i);
				}
				value = (T)sqrt((double)value);
			}
			return value;
		}
		GPL_Boolean IsBetween(SPA_Vector<S, T> &v1, SPA_Vector<S, T> &v2)
		{
			GPL_Integer i;
			for(i = 0;i < S;i++)
			{
				if(v1.Element(i) > v2.Element(i))
				{
					if(Element(i) >= v1.Element(i)) return FALSE;
					if(Element(i) <= v2.Element(i)) return FALSE;
				}
				else
				{
					if(Element(i) >= v2.Element(i)) return FALSE;
					if(Element(i) <= v1.Element(i)) return FALSE;
				}
			}
			return TRUE;
		}
		GPL_Boolean Swallows(T &t)
		{
			GPL_Integer i;
			for(i = 0;i < S;i++)
			{
				if(Element(i) < t){ return FALSE; }
			}
			return TRUE;
		}
		T Factorial()
		{
			GPL_Integer i;
			T t;
			t = Element(0);
			for(i = 1;i < S;i++)
			{
				t *= Element(i);
			}
			return t;
		}
		GPL_Integer GetSize()
		{
			return S;
		}
		GPL_String &String()
		{
			static GPL_String s;
			s.Clear();
			GPL_String n;
			GPL_Integer i;
			for(i = 0;i < S;i++)
			{
				if(i){s.Append("|");}
				n.Printf("%f", (double)(Element(i)));
				s.Append(n);
			}
			return s;
		}
};

/******************************************************************************
class SPA_RowVector
******************************************************************************/
template <GPL_Integer S, class T>
class SPA_RowVector : public SPA_Matrix<1, S, T>, public SPA_Vector<S, T>
{
	public:
		SPA_RowVector(void) {}
		virtual	~SPA_RowVector(void) {}

		T &operator [](const GPL_Integer index)
		{
			//return Element(index);
			return data[0][index];
		}
		T &Element(const GPL_Integer index)
		{
			//return SPA_Matrix<1, S, T>::data[0][index];
			return this->SPA_Matrix<1, S, T>::Element(0,index);
		}
};

/******************************************************************************
class SPA_ColVector
******************************************************************************/
template <GPL_Integer S, class T>
class SPA_ColVector : public SPA_Matrix<S, 1, T>, public SPA_Vector<S, T>
{
	public:
		SPA_ColVector(void) {}
		virtual	~SPA_ColVector(void) {}

		T &operator [](const GPL_Integer index)
		{
			//return Element(index);
			return data[index][0];
		}
		T &Element(const GPL_Integer index)
		{
			//return SPA_Matrix<S, 1, T>::data[index][0];
			return this->SPA_Matrix<S, 1, T>::Element(index,0);
		}
};

#endif /* SPA_VECTOR_H */
