#include <spa/spa.h>

static void printBytes(const char *text, const GPL_Block & bytes)
{
	GPL_String string;
	if (!gplByteArrayPrint(bytes, string))
	{
		fprintf(stderr, "Unable to print bytes!\n");
	}
	else
	{
		fprintf(stderr, "%s: \"%s\"\n", text, string.GetData());
	}
}

static void testEncryption(const char *text, const char *encrypterClass)
{
	fprintf(stderr, "Clear		: \"%s\"\n", text);

	// Create a new encrypter.
	GPL_Pointer<SPA_Encrypter> encrypter =
		(SPA_Encrypter *) GPL_ObjectFactory::Instance()->Create(encrypterClass);

	// Populate a message.
	GPL_Message message;
	message.AppendString(text);
	fprintf(stderr, "Message		: \"%s\"\n",
			message.ToString().GetData());

	// Copy to a byte array.
	GPL_Block bytes;
	message.CopyBuffer((GPL_Array<GPL_Byte> &) bytes);
	printBytes("Bytes		", bytes);

	// Encrypt.
	if (!encrypter->Encrypt(bytes))
	{
		fprintf(stderr, "Unable to encrypt!\n");
		return;
	}
	printBytes("Encrypt		", bytes);

	// Decrypt.
	if (!encrypter->Decrypt(bytes))
	{
		fprintf(stderr, "Unable to decrypt!\n");
		return;
	}
	printBytes("Decrypt		", bytes);

	fprintf(stderr, "Clear		: \"%s\"\n", text);
}

int main(const int argc, const char **argv)
{
	// Dummies to force the linker to include the object code for these classes.
	SPA_BlockEncrypter be((const unsigned long) 0xC0A8005F);
	SPA_ObfuscationEncrypter oe((const unsigned long) 0xC0A8005F);
	for(register int i = 1; argc > i; ++i)
	{
		testEncryption(argv[i], "SPA_BlockEncrypter");
		testEncryption(argv[i], "SPA_ObfuscationEncrypter");
	}
}
