#ifndef DNA_CONNECT_H
#define DNA_CONNECT_H

/******************************************************************************

Coral Tree dna/connect.h
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <gpl/gpl.h>
#include <dna/binding.h>

/***********************************************************************
class DNA_ConnectionImp

Network-layer peer connection implementation.
***********************************************************************/
class DNA_ConnectionImp : public DNA_BindingImp
{
	public:
		DNA_ConnectionImp(void) {}
		virtual ~DNA_ConnectionImp(void) {}

		virtual GPL_Boolean Accept(DNA_ConnectionImp *i, DNA_Error & error)
		{
			i = i; error = error;
			return FALSE;
		}
		virtual GPL_Boolean Connect(DNA_Error & error)
		{
			error = error;
			return FALSE;
		}
		virtual GPL_Boolean Disconnect(DNA_Error & error)
		{
			error = error;
			return FALSE;
		}

		// GPL_Object
		GPL_OBJECT_ABSTRACT_INTERFACE(DNA_ConnectionImp);
};

/***********************************************************************
enum DNA_ConnectionState
***********************************************************************/
enum DNA_ConnectionState
{
	DNA_CONNECTION_VALID		=	DNA_BINDING_VALID,
	DNA_CONNECTION_OPEN			=	DNA_BINDING_OPEN,
	DNA_CONNECTION_BOUND		=	DNA_BINDING_BOUND,
	DNA_CONNECTION_CONNECTED	=	0x00000008
};

/***********************************************************************
class DNA_Connection

Network-layer peer connection.
***********************************************************************/
class DNA_Connection : public DNA_Binding
{
	public:
		DNA_Connection(void) {}
		DNA_Connection(const DNA_Connection & c) { Attach(c); }
		virtual ~DNA_Connection(void) {}

		DNA_Connection & operator=(const DNA_Connection & c) { Detach(); Attach(c); return *this; }
		GPL_Boolean operator==(const DNA_Connection & c) const { return 0 == Compare(c); }

		virtual GPL_Boolean IsConnected(void);
		virtual DNA_ConnectionImp *GetConnectionImp(void) const
		{
			return (DNA_ConnectionImp *) DNA_Bus::GetImp();
		}

		virtual GPL_Boolean Accept(DNA_Connection & c);
		virtual GPL_Boolean Connect(void);
		virtual GPL_Boolean Disconnect(void);
		virtual GPL_Boolean Read(void *buffer, const unsigned long size,
							unsigned long & length);
		virtual GPL_Boolean Read(void *buffer, const unsigned long size,
							const long timeout, unsigned long & length);
		virtual GPL_Boolean Write(void *buffer, const unsigned long size,
					unsigned long & length);

		// GPL_Depictable
		virtual GPL_Boolean Print(GPL_String & string,
								const GPL_Verbosity verbosity) const;

		// GPL_Object
		GPL_OBJECT_INTERFACE(DNA_Connection);

	protected:
		virtual void Detach(void);
};

#endif /* DNA_CONNECT_H */
