#ifndef DNA_RMFBUS_H
#define DNA_RMFBUS_H

/******************************************************************************

Coral Tree dna/rmfbus.h
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/



#include <dna/bus.h>

#ifdef DNA_RMF

#include "RMF_Common.hh"
#include "RMF_App.hh"
#include "RAMP.hh"

/******************************************************************************
class DNA_GroupPoint

******************************************************************************/
class DNA_GroupPoint : public DNA_Point
	{
		IP_HOST_ADDRESS	address;
		IP_PORT			port;

	public:
						DNA_GroupPoint(void)
							{ SetAddressAndPort(0,0); };
						DNA_GroupPoint(IP_HOST_ADDRESS a,IP_PORT p)
							{ SetAddressAndPort(a,p); };

virtual					~DNA_GroupPoint(void)	{};

		void			SetAddress(IP_HOST_ADDRESS set)	{ address=set; };
		IP_HOST_ADDRESS	GetAddress(void)				{ return address; };

		void			SetPort(IP_PORT set)			{ port=set; };
		IP_PORT			GetPort(void)					{ return port; };

		void			SetAddressAndPort(IP_HOST_ADDRESS a,IP_PORT p)
							{
							SetAddress(a);
							SetPort(p);
							};

		DNA_GroupPoint	& operator=(const DNA_GroupPoint & other)	
							{
							SetAddressAndPort(other.address,other.port);
							return *this;
							};

		GPL_Boolean		operator==(const DNA_GroupPoint & other) const	
							{
							return (address==other.address && port==other.port);
							};
	};



/******************************************************************************
class	DNA_Group_BusImp : public DNA_BusImp

******************************************************************************/
class DNA_Group_BusImp : public DNA_BusImp
	{
	private:

		DNA_GroupPoint	point;
		DNA_GroupPoint	lastpoint;

	public:
						DNA_Group_BusImp(void)	{};
virtual					~DNA_Group_BusImp(void)	{};


		void			SetGroupPoint(const DNA_GroupPoint & set)
							{
							point=set;
							Reset();
							};
		DNA_GroupPoint	& GetGroupPoint(void)
							{ return point; };

		DNA_GroupPoint	& GetLastSendersPoint(void)	{ return lastpoint; };
		void			SetLastSendersPoint(DNA_GroupPoint & set)
							{
							lastpoint=set;
							};
		void			SetLastSendersPoint(IP_HOST_ADDRESS a,IP_PORT p)
							{
							lastpoint.SetAddressAndPort(a,p);
							};
		GPL_OBJECT_ABSTRACT_INTERFACE(DNA_Group_BusImp);
	};


/* DNA_RMF_GroupMember flags */
#define DNA_RMF_SELF		1

#if 0
class DNA_RMF_GroupMember
{
	private:
		UINT32			id;
		DNA_GroupPoint	point;
		unsigned long	flags;
	public:
		DNA_RMF_GroupMember(UINT32 receiver_id, DNA_RMF_Group_BusImp *local)
		{
			flags = 0;
			id = receiver_id;
			RMF_Address addr;
			if(RMF_OK != addr = local->GetSender()->GetAddressFromID(id))
			{
				wbaLog("DNA ERROR", "invalid receiver id");
				return;
			}
			if(id == local->GetReceiver()->GetID())
			{
				flags |= DNA_RMF_SELF;
			}
			point.SetAddress(addr.GetAddress());
			point.SetPort(addr.GetPort());
		}
virtual	~DNA_RMF_GroupMember()
		{
		}
		GPL_Boolean IsSelf(void) { return flags & DNA_RMF_SELF; }
};
#endif

class DNA_RMF_Config
{
	public:
		INT32	reliabilityMode;
		INT32	rampWindowSize;
		DNA_RMF_Config()
		{
			reliabilityMode = RMF_NAK_MODE;
			rampWindowSize = 32;
		}
};

/******************************************************************************
class	DNA_RMF_Group_BusImp : public DNA_Group_BusImp

******************************************************************************/
class DNA_RMF_Group_BusImp : public DNA_Group_BusImp
	{
	private:
		RAMP									*sender;
		RMF_UniversalReceiver					*receiver;
		long									opened;

	public:
					DNA_RMF_Group_BusImp(void);
virtual				~DNA_RMF_Group_BusImp(void);

		RAMP		*GetSender(){return sender;}
		RMF_UniversalReceiver *GetReceiver(){return receiver;}

virtual	long		Open(DNA_Error & error);
virtual	long		Open(DNA_Error & error, DNA_RMF_Config &config);
virtual	void		Close(DNA_Error & error);
virtual	void		Reset(void);

virtual	long		Compare(const DNA_BusImp *imp) const;

virtual	GPL_Boolean	Read(void *buffer, const unsigned long size,
									unsigned long & length, DNA_Error & error);

virtual	GPL_Boolean	Read(void *buffer, const unsigned long size,
									const long timeout, unsigned long & length,
									DNA_Error & error);

virtual	GPL_Boolean	Write(void *buffer, const unsigned long size,
									unsigned long & length, DNA_Error & error);


		GPL_Boolean	IsOpen(void) { return opened; }
		GPL_OBJECT_INTERFACE(DNA_RMF_Group_BusImp);
	};

#endif /* DNA_RMF */

#endif /* DNA_RMFBUS_H */
