
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <dna/dna.h>

/******************************************************************************
DNA_Server::DNA_Server(void)
******************************************************************************/
DNA_Server::DNA_Server(void) : connections(HashFunction)
{
	// Provide a default implementation of TCP/IP.
	connection.SetImp((DNA_BusImp *) new DNA_TCPConnectionImp());
}

/******************************************************************************
GPL_Boolean DNA_Server::Initialize(void)
******************************************************************************/
GPL_Boolean DNA_Server::Initialize(void)
{
	// This presumes that a sub-class has specified the source point.
	register DNA_TCPConnectionImp *imp =
		(DNA_TCPConnectionImp *) connection.GetConnectionImp();
	if (!connection.Open())
	{
		return FALSE;
	}
	if (!connection.Bind())
	{
		return FALSE;
	}
	imp->SetBackLog(5);

	return TRUE;
}

/******************************************************************************
void DNA_Server::Finalize(void)
******************************************************************************/
void DNA_Server::Finalize(void)
{
	// Explicitly disconnect and close all connections.
	GPL_MapIterator<SOCKET_HANDLE, DNA_Connection> iterator;
	iterator.Reset(&connections);
	while (iterator.Next())
	{
		// Ignore the server conection.
		if (DNA_GET_HANDLE(connection) == iterator.Key())
		{
			iterator.Value().Disconnect();
		}
		iterator.Value().Close();
	}
}

/***********************************************************************
GPL_Boolean DNA_Server::Request(DNA_Message & message, const GPL_Opaque key)
***********************************************************************/
GPL_Boolean DNA_Server::Request(DNA_Message & message, const GPL_Opaque key)
{
	// Validate the key.
	const SOCKET_HANDLE handle = (SOCKET_HANDLE) key;
	if (connections.Contains(handle))
	{
		// Send the reply.
		const GPL_String image = message.ToString();
		wbaLog("DNA LOG", "SND >%s<", image.GetData());
		return message.Send(&(connections[handle]));
	}
	return FALSE;
}

/***********************************************************************
GPL_Boolean DNA_Server::Reply(DNA_Message & message, const GPL_Opaque key)
***********************************************************************/
GPL_Boolean DNA_Server::Reply(DNA_Message & message, const GPL_Opaque key)
{
	// Validate the key.
	const SOCKET_HANDLE handle = (SOCKET_HANDLE) key;
	if (connections.Contains(handle))
	{
		// Send the reply.
		const GPL_String image = message.ToString();
		wbaLog("DNA LOG", "SND >%s<", image.GetData());
		return message.Send(&(connections[handle]));
	}
	return FALSE;
}

/***********************************************************************
void DNA_Server::Reject(const GPL_Opaque key)
***********************************************************************/
void DNA_Server::Reject(const GPL_Opaque key)
{
	// Validate the key.
	const SOCKET_HANDLE handle = (SOCKET_HANDLE) key;
	if (connections.Contains(handle))
	{
		// Disconnect from the client.
		connections[handle].Disconnect();

		// Close the connection.
		connections[handle].Close();

		// Remove from map
		connections.Remove(handle);
	}
}

